# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is Mozilla-specific Buildbot steps.
#
# The Initial Developer of the Original Code is
# Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2009
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Brian Warner <warner@lothar.com>
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

import os.path

from zope.interface import implements
from twisted.application import strports
from twisted.python import log, runtime
from twisted.protocols import basic
from twisted.cred import portal, checkers
from twisted.spread import pb

from buildbot import pbutil
from buildbot.sourcestamp import SourceStamp
from buildbot.changes.maildir import MaildirService
from buildbot.process.properties import Properties
from buildbot.schedulers import base
from buildbot.status.builder import BuildSetStatus


class TryBase(base.BaseScheduler):

    def run(self):
        # triggered by external events, not DB changes or timers
        return None

    def filterBuilderList(self, builderNames):
        # self.builderNames is the configured list of builders
        # available for try.  If the user supplies a list of builders,
        # it must be restricted to the configured list.  If not, build
        # on all of the configured builders.
        if builderNames:
            for b in builderNames:
                if not b in self.builderNames:
                    log.msg("%s got with builder %s" % (self, b))
                    log.msg(" but that wasn't in our list: %s"
                            % (self.builderNames,))
                    return []
        else:
            builderNames = self.builderNames
        return builderNames

class BadJobfile(Exception):
    pass

class JobFileScanner(basic.NetstringReceiver):
    def __init__(self):
        self.strings = []
        self.transport = self # so transport.loseConnection works
        self.error = False

    def stringReceived(self, s):
        self.strings.append(s)

    def loseConnection(self):
        self.error = True

class Try_Jobdir(TryBase):
    compare_attrs = ( 'name', 'builderNames', 'jobdir', 'properties' )

    def __init__(self, name, builderNames, jobdir,
                 properties={}):
        base.BaseScheduler.__init__(self, name, builderNames, properties)
        self.jobdir = jobdir
        self.watcher = MaildirService()
        self.watcher.setServiceParent(self)

    def setServiceParent(self, parent):
        sm = parent
        m = sm.parent
        self.watcher.setBasedir(os.path.join(m.basedir, self.jobdir))
        TryBase.setServiceParent(self, parent)

    def parseJob(self, f):
        # jobfiles are serialized build requests. Each is a list of
        # serialized netstrings, in the following order:
        #  "1", the version number of this format
        #  buildsetID, arbitrary string, used to find the buildSet later
        #  branch name, "" for default-branch
        #  base revision, "" for HEAD
        #  patchlevel, usually "1"
        #  patch
        #  builderNames...
        p = JobFileScanner()
        p.dataReceived(f.read())
        if p.error:
            raise BadJobfile("unable to parse netstrings")
        s = p.strings
        ver = s.pop(0)
        if ver == "1":
            buildsetID, branch, baserev, patchlevel, diff = s[:5]
            builderNames = s[5:]
            if branch == "":
                branch = None
            if baserev == "":
                baserev = None
            patchlevel = int(patchlevel)
            patch = (patchlevel, diff)
            ss = SourceStamp("Old client", branch, baserev, patch)
        elif ver == "2": # introduced the repository and project property
            buildsetID, branch, baserev, patchlevel, diff, repository, project = s[:7]
            builderNames = s[7:]
            if branch == "":
                branch = None
            if baserev == "":
                baserev = None
            patchlevel = int(patchlevel)
            patch = (patchlevel, diff)
            ss = SourceStamp(branch, baserev, patch, repository=repository,
                             project=project)
        else:
            raise BadJobfile("unknown version '%s'" % ver)
        return builderNames, ss, buildsetID

    def messageReceived(self, filename):
        md = os.path.join(self.parent.parent.basedir, self.jobdir)
        if runtime.platformType == "posix":
            # open the file before moving it, because I'm afraid that once
            # it's in cur/, someone might delete it at any moment
            path = os.path.join(md, "new", filename)
            f = open(path, "r")
            os.rename(os.path.join(md, "new", filename),
                      os.path.join(md, "cur", filename))
        else:
            # do this backwards under windows, because you can't move a file
            # that somebody is holding open. This was causing a Permission
            # Denied error on bear's win32-twisted1.3 buildslave.
            os.rename(os.path.join(md, "new", filename),
                      os.path.join(md, "cur", filename))
            path = os.path.join(md, "cur", filename)
            f = open(path, "r")

        try:
            builderNames, ss, jobid = self.parseJob(f)
        except BadJobfile:
            log.msg("%s reports a bad jobfile in %s" % (self, filename))
            log.err()
            return
        # Validate/fixup the builder names.
        builderNames = self.filterBuilderList(builderNames)
        if not builderNames:
            return
        reason = "'try' job"
        d = self.parent.db.runInteraction(self._try, ss, builderNames, reason)
        def _done(ign):
            self.parent.loop_done() # so it will notify builder loop
        d.addCallback(_done)
        return d

    def _try(self, t, ss, builderNames, reason):
        db = self.parent.db
        ssid = db.get_sourcestampid(ss, t)
        bsid = self.create_buildset(ssid, reason, t, builderNames=builderNames)
        return bsid

class Try_Userpass(TryBase):
    compare_attrs = ( 'name', 'builderNames', 'port', 'userpass', 'properties' )
    implements(portal.IRealm)

    def __init__(self, name, builderNames, port, userpass,
                 properties={}):
        base.BaseScheduler.__init__(self, name, builderNames, properties)
        if type(port) is int:
            port = "tcp:%d" % port
        self.port = port
        self.userpass = userpass
        c = checkers.InMemoryUsernamePasswordDatabaseDontUse()
        for user,passwd in self.userpass:
            c.addUser(user, passwd)

        p = portal.Portal(self)
        p.registerChecker(c)
        f = pb.PBServerFactory(p)
        s = strports.service(port, f)
        s.setServiceParent(self)

    def getPort(self):
        # utility method for tests: figure out which TCP port we just opened.
        return self.services[0]._port.getHost().port

    def requestAvatar(self, avatarID, mind, interface):
        log.msg("%s got connection from user %s" % (self, avatarID))
        assert interface == pb.IPerspective
        p = Try_Userpass_Perspective(self, avatarID)
        return (pb.IPerspective, p, lambda: None)

class Try_Userpass_Perspective(pbutil.NewCredPerspective):
    def __init__(self, parent, username):
        self.parent = parent
        self.username = username

    def perspective_try(self, branch, revision, patch, repository, project,
                        builderNames, properties={}, ):
        log.msg("user %s requesting build on builders %s" % (self.username,
                                                             builderNames))
        # build the intersection of the request and our configured list
        builderNames = self.parent.filterBuilderList(builderNames)
        if not builderNames:
            return
        ss = SourceStamp(branch, revision, patch, repository=repository,
                         project=project)
        reason = "'try' job from user %s" % self.username

        # roll the specified props in with our inherited props
        combined_props = Properties()
        combined_props.updateFromProperties(self.parent.properties)
        combined_props.update(properties, "try build")

        status = self.parent.parent.parent.status
        db = self.parent.parent.db
        d = db.runInteraction(self._try, ss, builderNames, reason,
                              combined_props, db)
        def _done(bsid):
            # return a remotely-usable BuildSetStatus object
            bss = BuildSetStatus(bsid, status, db)
            from buildbot.status.client import makeRemote
            r = makeRemote(bss)
            #self.parent.parent.loop_done() # so it will notify builder loop
            return r
        d.addCallback(_done)
        return d

    def _try(self, t, ss, builderNames, reason, combined_props, db):
        ssid = db.get_sourcestampid(ss, t)
        bsid = self.parent.create_buildset(ssid, reason, t,
                                           props=combined_props,
                                           builderNames=builderNames)
        return bsid

    def perspective_getAvailableBuilderNames(self):
        # Return a list of builder names that are configured
        # for the try service
        # This is mostly intended for integrating try services
        # into other applications
        return self.parent.listBuilderNames()



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is Mozilla-specific Buildbot steps.
#
# The Initial Developer of the Original Code is
# Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2009
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Brian Warner <warner@lothar.com>
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

import time
from twisted.internet import defer
from twisted.python import log
from buildbot.sourcestamp import SourceStamp
from buildbot.schedulers import base

class TimedBuildMixin:

    def start_HEAD_build(self, t):
        # start a build (of the tip of self.branch)
        db = self.parent.db
        ss = SourceStamp(branch=self.branch)
        ssid = db.get_sourcestampid(ss, t)
        self.create_buildset(ssid, self.reason, t)

    def start_requested_build(self, t, relevant_changes):
        # start a build with the requested list of changes on self.branch
        db = self.parent.db
        ss = SourceStamp(branch=self.branch, changes=relevant_changes)
        ssid = db.get_sourcestampid(ss, t)
        self.create_buildset(ssid, self.reason, t)

    def update_last_build(self, t, when):
        # and record when we did it
        state = self.get_state(t)
        state["last_build"] = when
        self.set_state(t, state)

class Periodic(base.BaseScheduler, TimedBuildMixin):
    """Instead of watching for Changes, this Scheduler can just start a build
    at fixed intervals. The C{periodicBuildTimer} parameter sets the number
    of seconds to wait between such periodic builds. The first build will be
    run immediately."""

    # TODO: consider having this watch another (changed-based) scheduler and
    # merely enforce a minimum time between builds.
    compare_attrs = ('name', 'builderNames', 'periodicBuildTimer', 'branch',
                     'properties')

    def __init__(self, name, builderNames, periodicBuildTimer,
            branch=None, properties={}):
        base.BaseScheduler.__init__(self, name, builderNames, properties)
        self.periodicBuildTimer = periodicBuildTimer
        self.branch = branch
        self.reason = ("The Periodic scheduler named '%s' triggered this build"
                       % name)

    def get_initial_state(self, max_changeid):
        return {"last_build": None}

    def getPendingBuildTimes(self):
        db = self.parent.db
        s = db.runInteractionNow(self.get_state)
        last_build = s["last_build"]
        now = time.time()
        if last_build is None:
            return [now]
        return [last_build + self.periodicBuildTimer]

    def run(self):
        db = self.parent.db
        d = db.runInteraction(self._run)
        return d

    def _run(self, t):
        now = time.time()
        s = self.get_state(t)
        last_build = s["last_build"]
        if last_build is None:
            self.start_HEAD_build(t)
            self.update_last_build(t, now)
            last_build = now
        when = last_build + self.periodicBuildTimer
        if when < now:
            self.start_HEAD_build(t)
            self.update_last_build(t, now)
            last_build = now
            when = now + self.periodicBuildTimer
        return when + 1.0


class Nightly(base.BaseScheduler, base.ClassifierMixin, TimedBuildMixin):
    """Imitate 'cron' scheduling. This can be used to schedule a nightly
    build, or one which runs are certain times of the day, week, or month.

    Pass some subset of minute, hour, dayOfMonth, month, and dayOfWeek; each
    may be a single number or a list of valid values. The builds will be
    triggered whenever the current time matches these values. Wildcards are
    represented by a '*' string. All fields default to a wildcard except
    'minute', so with no fields this defaults to a build every hour, on the
    hour.

    For example, the following master.cfg clause will cause a build to be
    started every night at 3:00am::

     s = Nightly(name='nightly', builderNames=['builder1', 'builder2'],
                 hour=3, minute=0)
     c['schedules'].append(s)

    This scheduler will perform a build each monday morning at 6:23am and
    again at 8:23am::

     s = Nightly(name='BeforeWork', builderNames=['builder1'],
                 dayOfWeek=0, hour=[6,8], minute=23)

    The following runs a build every two hours::

     s = Nightly(name='every2hours', builderNames=['builder1'],
                 hour=range(0, 24, 2))

    And this one will run only on December 24th::

     s = Nightly(name='SleighPreflightCheck',
                 builderNames=['flying_circuits', 'radar'],
                 month=12, dayOfMonth=24, hour=12, minute=0)

    For dayOfWeek and dayOfMonth, builds are triggered if the date matches
    either of them. All time values are compared against the tuple returned
    by time.localtime(), so month and dayOfMonth numbers start at 1, not
    zero. dayOfWeek=0 is Monday, dayOfWeek=6 is Sunday.

    When onlyIfChanged is True, the build is triggered only if changes have
    arrived on the given branch since the last build was performed. As a
    further restriction, if fileIsImportant= is provided (a one-argument
    callable which takes a Change object and returns a bool), then the build
    will be triggered only if at least one of those changes qualifies as
    'important'. The following example will run a build at 3am, but only when
    a source code file (.c/.h) has been changed:

     def isSourceFile(change):
         for fn in change.files:
             if fn.endswith('.c') or fn.endswith('.h'):
                 return True
         return False
     s = Nightly(name='nightly-when-changed', builderNames=['builder1'],
                 hour=3, minute=0,
                 onlyIfChanged=True, fileIsImportant=isSourceFile)

    onlyIfChanged defaults to False, which means a build will be performed
    even if nothing has changed.
    """

    compare_attrs = ('name', 'builderNames',
                     'minute', 'hour', 'dayOfMonth', 'month',
                     'dayOfWeek', 'branch', 'onlyIfChanged',
                     'fileIsImportant', 'properties')

    def __init__(self, name, builderNames, minute=0, hour='*',
                 dayOfMonth='*', month='*', dayOfWeek='*',
                 branch=None, fileIsImportant=None, onlyIfChanged=False,
                 properties={}):
        # Setting minute=0 really makes this an 'Hourly' scheduler. This
        # seemed like a better default than minute='*', which would result in
        # a build every 60 seconds.
        base.BaseScheduler.__init__(self, name, builderNames, properties)
        self.minute = minute
        self.hour = hour
        self.dayOfMonth = dayOfMonth
        self.month = month
        self.dayOfWeek = dayOfWeek
        self.branch = branch
        self.onlyIfChanged = onlyIfChanged
        self.delayedRun = None
        self.nextRunTime = None
        self.reason = ("The Nightly scheduler named '%s' triggered this build"
                       % name)
        self.fileIsImportant = None
        if fileIsImportant:
            assert callable(fileIsImportant)
            self.fileIsImportant = fileIsImportant
        self._start_time = time.time()

        # this scheduler does not support filtering, but ClassifierMixin needs a
        # filter anyway
        self.make_filter()

    def get_initial_state(self, max_changeid):
        return {
            "last_build": None,
            "last_processed": max_changeid,
        }

    def getPendingBuildTimes(self):
        now = time.time()
        next = self._calculateNextRunTimeFrom(now)
        # note: this ignores onlyIfChanged
        return [next]

    def run(self):
        d = defer.succeed(None)
        db = self.parent.db
        # always call classify_changes, so that we can keep last_processed
        # up to date, in case we are configured with onlyIfChanged.
        d.addCallback(lambda ign: db.runInteraction(self.classify_changes))
        d.addCallback(lambda ign: db.runInteraction(self._check_timer))
        return d

    def _check_timer(self, t):
        now = time.time()
        s = self.get_state(t)
        last_build = s["last_build"]
        if last_build is None:
            next = self._calculateNextRunTimeFrom(self._start_time)
        else:
            next = self._calculateNextRunTimeFrom(last_build)

        # not ready to fire yet
        if next >= now:
            return next + 1.0

        self._maybe_start_build(t)
        self.update_last_build(t, now)

        # reschedule for the next timer
        return self._check_timer(t)

    def _maybe_start_build(self, t):
        if self.onlyIfChanged:
            db = self.parent.db
            res = db.scheduler_get_classified_changes(self.schedulerid, t)
            (important, unimportant) = res
            if not important:
                log.msg("Nightly Scheduler <%s>: "
                        "skipping build - No important change" % self.name)
                return
            relevant_changes = [c for c in (important + unimportant) if
                                c.branch == self.branch]
            if not relevant_changes:
                log.msg("Nightly Scheduler <%s>: "
                        "skipping build - No relevant change on branch" %
                        self.name)
                return
            self.start_requested_build(t, relevant_changes)
            # retire the changes
            changeids = [c.number for c in relevant_changes]
            db.scheduler_retire_changes(self.schedulerid, changeids, t)
        else:
            # start it unconditionally
            self.start_HEAD_build(t)

    def _addTime(self, timetuple, secs):
        return time.localtime(time.mktime(timetuple)+secs)

    def _isRunTime(self, timetuple):
        def check(ourvalue, value):
            if ourvalue == '*': return True
            if isinstance(ourvalue, int): return value == ourvalue
            return (value in ourvalue)

        if not check(self.minute, timetuple[4]):
            #print 'bad minute', timetuple[4], self.minute
            return False

        if not check(self.hour, timetuple[3]):
            #print 'bad hour', timetuple[3], self.hour
            return False

        if not check(self.month, timetuple[1]):
            #print 'bad month', timetuple[1], self.month
            return False

        if self.dayOfMonth != '*' and self.dayOfWeek != '*':
            # They specified both day(s) of month AND day(s) of week.
            # This means that we only have to match one of the two. If
            # neither one matches, this time is not the right time.
            if not (check(self.dayOfMonth, timetuple[2]) or
                    check(self.dayOfWeek, timetuple[6])):
                #print 'bad day'
                return False
        else:
            if not check(self.dayOfMonth, timetuple[2]):
                #print 'bad day of month'
                return False

            if not check(self.dayOfWeek, timetuple[6]):
                #print 'bad day of week'
                return False

        return True

    def _calculateNextRunTimeFrom(self, now):
        dateTime = time.localtime(now)

        # Remove seconds by advancing to at least the next minute
        dateTime = self._addTime(dateTime, 60-dateTime[5])

        # Now we just keep adding minutes until we find something that matches

        # It not an efficient algorithm, but it'll *work* for now
        yearLimit = dateTime[0]+2
        while not self._isRunTime(dateTime):
            dateTime = self._addTime(dateTime, 60)
            #print 'Trying', time.asctime(dateTime)
            assert dateTime[0] < yearLimit, 'Something is wrong with this code'
        return time.mktime(dateTime)



# -*- test-case-name: buildbot.test.test_run -*-

import os
import signal
import time
import textwrap

from zope.interface import implements
from twisted.python import log, components
from twisted.python.failure import Failure
from twisted.internet import defer, reactor
from twisted.spread import pb
from twisted.cred import portal, checkers
from twisted.application import service, strports
from twisted.application.internet import TimerService

import buildbot
# sibling imports
from buildbot.util import now, safeTranslate, eventual
from buildbot.pbutil import NewCredPerspective
from buildbot.process.builder import Builder, IDLE
from buildbot.status.builder import Status, BuildSetStatus
from buildbot.changes.changes import Change
from buildbot.changes.manager import ChangeManager
from buildbot import interfaces, locks
from buildbot.process.properties import Properties
from buildbot.config import BuilderConfig
from buildbot.process.builder import BuilderControl
from buildbot.db.dbspec import DBSpec
from buildbot.db import connector, exceptions
from buildbot.db.schema.manager import DBSchemaManager
from buildbot.schedulers.manager import SchedulerManager
from buildbot.util.loop import DelegateLoop

########################################

class BotMaster(service.MultiService):

    """This is the master-side service which manages remote buildbot slaves.
    It provides them with BuildSlaves, and distributes file change
    notification messages to them.
    """

    debug = 0
    reactor = reactor

    def __init__(self):
        service.MultiService.__init__(self)
        self.builders = {}
        self.builderNames = []
        # builders maps Builder names to instances of bb.p.builder.Builder,
        # which is the master-side object that defines and controls a build.
        # They are added by calling botmaster.addBuilder() from the startup
        # code.

        # self.slaves contains a ready BuildSlave instance for each
        # potential buildslave, i.e. all the ones listed in the config file.
        # If the slave is connected, self.slaves[slavename].slave will
        # contain a RemoteReference to their Bot instance. If it is not
        # connected, that attribute will hold None.
        self.slaves = {} # maps slavename to BuildSlave
        self.statusClientService = None
        self.watchers = {}

        # self.locks holds the real Lock instances
        self.locks = {}

        # self.mergeRequests is the callable override for merging build
        # requests
        self.mergeRequests = None

        # self.prioritizeBuilders is the callable override for builder order
        # traversal
        self.prioritizeBuilders = None

        self.loop = DelegateLoop(self._get_processors)
        self.loop.setServiceParent(self)

        self.shuttingDown = False

    def setMasterName(self, name, incarnation):
        self.master_name = name
        self.master_incarnation = incarnation

    def cleanShutdown(self):
        if self.shuttingDown:
            return
        log.msg("Initiating clean shutdown")
        self.shuttingDown = True

        # Wait for all builds to finish
        l = []
        for builder in self.builders.values():
            for build in builder.builder_status.getCurrentBuilds():
                l.append(build.waitUntilFinished())
        if len(l) == 0:
            log.msg("No running jobs, starting shutdown immediately")
            self.loop.trigger()
            d = self.loop.when_quiet()
        else:
            log.msg("Waiting for %i build(s) to finish" % len(l))
            d = defer.DeferredList(l)
            d.addCallback(lambda ign: self.loop.when_quiet())

        # Flush the eventual queue
        d.addCallback(eventual.flushEventualQueue)

        # Finally, shut the whole process down
        def shutdown(ign):
            # Double check that we're still supposed to be shutting down
            # The shutdown may have been cancelled!
            if self.shuttingDown:
                # Check that there really aren't any running builds
                for builder in self.builders.values():
                    n = len(builder.builder_status.getCurrentBuilds())
                    if n > 0:
                        log.msg("Not shutting down, builder %s has %i builds running" % (builder, n))
                        log.msg("Trying shutdown sequence again")
                        self.shuttingDown = False
                        self.cleanShutdown()
                        return
                log.msg("Stopping reactor")
                self.reactor.stop()
        d.addCallback(shutdown)
        return d

    def cancelCleanShutdown(self):
        if not self.shuttingDown:
            return
        log.msg("Cancelling clean shutdown")
        self.shuttingDown = False

    def _sortfunc(self, b1, b2):
        t1 = b1.getOldestRequestTime()
        t2 = b2.getOldestRequestTime()
        # If t1 or t2 is None, then there are no build requests,
        # so sort it at the end
        if t1 is None:
            return 1
        if t2 is None:
            return -1
        return cmp(t1, t2)

    def _sort_builders(self, parent, builders):
        return sorted(builders, self._sortfunc)

    def _get_processors(self):
        if self.shuttingDown:
            return []
        builders = self.builders.values()
        sorter = self.prioritizeBuilders or self._sort_builders
        try:
            builders = sorter(self.parent, builders)
        except:
            log.msg("Exception prioritizing builders")
            log.err(Failure())
            # leave them in the original order
        return [b.run for b in builders]

    def trigger_add_buildrequest(self, category, *brids):
        # a buildrequest has been added or resubmitted
        self.loop.trigger()
    def triggerNewBuildCheck(self):
        # called when a build finishes, or a slave attaches
        self.loop.trigger()

    # these four are convenience functions for testing

    def waitUntilBuilderAttached(self, name):
        b = self.builders[name]
        #if b.slaves:
        #    return defer.succeed(None)
        d = defer.Deferred()
        b.watchers['attach'].append(d)
        return d

    def waitUntilBuilderDetached(self, name):
        b = self.builders.get(name)
        if not b or not b.slaves:
            return defer.succeed(None)
        d = defer.Deferred()
        b.watchers['detach'].append(d)
        return d

    def waitUntilBuilderFullyDetached(self, name):
        b = self.builders.get(name)
        # TODO: this looks too deeply inside the Builder object
        if not b or not b.slaves:
            return defer.succeed(None)
        d = defer.Deferred()
        b.watchers['detach_all'].append(d)
        return d

    def waitUntilBuilderIdle(self, name):
        b = self.builders[name]
        # TODO: this looks way too deeply inside the Builder object
        for sb in b.slaves:
            if sb.state != IDLE:
                d = defer.Deferred()
                b.watchers['idle'].append(d)
                return d
        return defer.succeed(None)

    def loadConfig_Slaves(self, new_slaves):
        old_slaves = [c for c in list(self)
                      if interfaces.IBuildSlave.providedBy(c)]

        # identify added/removed slaves. For each slave we construct a tuple
        # of (name, password, class), and we consider the slave to be already
        # present if the tuples match. (we include the class to make sure
        # that BuildSlave(name,pw) is different than
        # SubclassOfBuildSlave(name,pw) ). If the password or class has
        # changed, we will remove the old version of the slave and replace it
        # with a new one. If anything else has changed, we just update the
        # old BuildSlave instance in place. If the name has changed, of
        # course, it looks exactly the same as deleting one slave and adding
        # an unrelated one.
        old_t = {}
        for s in old_slaves:
            old_t[(s.slavename, s.password, s.__class__)] = s
        new_t = {}
        for s in new_slaves:
            new_t[(s.slavename, s.password, s.__class__)] = s
        removed = [old_t[t]
                   for t in old_t
                   if t not in new_t]
        added = [new_t[t]
                 for t in new_t
                 if t not in old_t]
        remaining_t = [t
                       for t in new_t
                       if t in old_t]
        # removeSlave will hang up on the old bot
        dl = []
        for s in removed:
            dl.append(self.removeSlave(s))
        d = defer.DeferredList(dl, fireOnOneErrback=True)
        def _add(res):
            for s in added:
                self.addSlave(s)
            for t in remaining_t:
                old_t[t].update(new_t[t])
        d.addCallback(_add)
        return d

    def addSlave(self, s):
        s.setServiceParent(self)
        s.setBotmaster(self)
        self.slaves[s.slavename] = s

    def removeSlave(self, s):
        # TODO: technically, disownServiceParent could return a Deferred
        s.disownServiceParent()
        d = self.slaves[s.slavename].disconnect()
        del self.slaves[s.slavename]
        return d

    def slaveLost(self, bot):
        for name, b in self.builders.items():
            if bot.slavename in b.slavenames:
                b.detached(bot)

    def getBuildersForSlave(self, slavename):
        return [b
                for b in self.builders.values()
                if slavename in b.slavenames]

    def getBuildernames(self):
        return self.builderNames

    def getBuilders(self):
        allBuilders = [self.builders[name] for name in self.builderNames]
        return allBuilders

    def setBuilders(self, builders):
        # TODO: remove self.builders and just use the Service hierarchy to
        # keep track of active builders. We could keep self.builderNames to
        # retain ordering, if it seems important.
        self.builders = {}
        self.builderNames = []
        d = defer.DeferredList([b.disownServiceParent() for b in list(self)
                                if isinstance(b, Builder)],
                               fireOnOneErrback=True)
        def _add(ign):
            log.msg("setBuilders._add: %s %s" % (list(self), builders))
            for b in builders:
                for slavename in b.slavenames:
                    # this is actually validated earlier
                    assert slavename in self.slaves
                self.builders[b.name] = b
                self.builderNames.append(b.name)
                b.setBotmaster(self)
                b.setServiceParent(self)
        d.addCallback(_add)
        d.addCallback(lambda ign: self._updateAllSlaves())
        return d

    def _updateAllSlaves(self):
        """Notify all buildslaves about changes in their Builders."""
        dl = []
        for s in self.slaves.values():
            d = s.updateSlave()
            d.addErrback(log.err)
            dl.append(d)
        return defer.DeferredList(dl)

    def shouldMergeRequests(self, builder, req1, req2):
        """Determine whether two BuildRequests should be merged for
        the given builder.

        """
        if self.mergeRequests is not None:
            return self.mergeRequests(builder, req1, req2)
        return req1.canBeMergedWith(req2)

    def getPerspective(self, slavename):
        return self.slaves[slavename]

    def shutdownSlaves(self):
        # TODO: make this into a bot method rather than a builder method
        for b in self.slaves.values():
            b.shutdownSlave()

    def stopService(self):
        for b in self.builders.values():
            b.builder_status.addPointEvent(["master", "shutdown"])
            b.builder_status.saveYourself()
        return service.MultiService.stopService(self)

    def getLockByID(self, lockid):
        """Convert a Lock identifier into an actual Lock instance.
        @param lockid: a locks.MasterLock or locks.SlaveLock instance
        @return: a locks.RealMasterLock or locks.RealSlaveLock instance
        """
        assert isinstance(lockid, (locks.MasterLock, locks.SlaveLock))
        if not lockid in self.locks:
            self.locks[lockid] = lockid.lockClass(lockid)
        # if the master.cfg file has changed maxCount= on the lock, the next
        # time a build is started, they'll get a new RealLock instance. Note
        # that this requires that MasterLock and SlaveLock (marker) instances
        # be hashable and that they should compare properly.
        return self.locks[lockid]

########################################



class DebugPerspective(NewCredPerspective):
    def attached(self, mind):
        return self
    def detached(self, mind):
        pass

    def perspective_pingBuilder(self, buildername):
        c = interfaces.IControl(self.master)
        bc = c.getBuilder(buildername)
        bc.ping()

    def perspective_fakeChange(self, file, revision=None, who="fakeUser",
                               branch=None, repository="", 
                               project=""):
        change = Change(who, [file], "some fake comments\n",
                        branch=branch, revision=revision,
                        repository=repository, project=project)
        c = interfaces.IControl(self.master)
        c.addChange(change)

    def perspective_setCurrentState(self, buildername, state):
        builder = self.botmaster.builders.get(buildername)
        if not builder: return
        if state == "offline":
            builder.statusbag.currentlyOffline()
        if state == "idle":
            builder.statusbag.currentlyIdle()
        if state == "waiting":
            builder.statusbag.currentlyWaiting(now()+10)
        if state == "building":
            builder.statusbag.currentlyBuilding(None)
    def perspective_reload(self):
        print "doing reload of the config file"
        self.master.loadTheConfigFile()
    def perspective_pokeIRC(self):
        print "saying something on IRC"
        from buildbot.status import words
        for s in self.master:
            if isinstance(s, words.IRC):
                bot = s.f
                for channel in bot.channels:
                    print " channel", channel
                    bot.p.msg(channel, "Ow, quit it")

    def perspective_print(self, msg):
        print "debug", msg

class Dispatcher:
    implements(portal.IRealm)

    def __init__(self):
        self.names = {}

    def register(self, name, afactory):
        self.names[name] = afactory
    def unregister(self, name):
        del self.names[name]

    def requestAvatar(self, avatarID, mind, interface):
        assert interface == pb.IPerspective
        afactory = self.names.get(avatarID)
        if afactory:
            p = afactory.getPerspective()
        elif avatarID == "change":
            raise ValueError("no PBChangeSource installed")
        elif avatarID == "debug":
            p = DebugPerspective()
            p.master = self.master
            p.botmaster = self.botmaster
        elif avatarID == "statusClient":
            p = self.statusClientService.getPerspective()
        else:
            # it must be one of the buildslaves: no other names will make it
            # past the checker
            p = self.botmaster.getPerspective(avatarID)

        if not p:
            raise ValueError("no perspective for '%s'" % avatarID)

        d = defer.maybeDeferred(p.attached, mind)
        d.addCallback(self._avatarAttached, mind)
        return d

    def _avatarAttached(self, p, mind):
        return (pb.IPerspective, p, lambda p=p,mind=mind: p.detached(mind))

########################################

class _Unset: pass  # marker

class LogRotation: 
    '''holds log rotation parameters (for WebStatus)'''
    def __init__(self):
        self.rotateLength = 1 * 1000 * 1000 
        self.maxRotatedFiles = 10

class BuildMaster(service.MultiService):
    debug = 0
    manhole = None
    debugPassword = None
    projectName = "(unspecified)"
    projectURL = None
    buildbotURL = None
    change_svc = None
    properties = Properties()

    def __init__(self, basedir, configFileName="master.cfg", db_spec=None):
        service.MultiService.__init__(self)
        self.setName("buildmaster")
        self.basedir = basedir
        self.configFileName = configFileName

        # the dispatcher is the realm in which all inbound connections are
        # looked up: slave builders, change notifications, status clients, and
        # the debug port
        dispatcher = Dispatcher()
        dispatcher.master = self
        self.dispatcher = dispatcher
        self.checker = checkers.InMemoryUsernamePasswordDatabaseDontUse()
        # the checker starts with no user/passwd pairs: they are added later
        p = portal.Portal(dispatcher)
        p.registerChecker(self.checker)
        self.slaveFactory = pb.PBServerFactory(p)
        self.slaveFactory.unsafeTracebacks = True # let them see exceptions

        self.slavePortnum = None
        self.slavePort = None

        self.change_svc = ChangeManager()
        self.change_svc.setServiceParent(self)
        self.dispatcher.changemaster = self.change_svc

        try:
            hostname = os.uname()[1] # only on unix
        except AttributeError:
            hostname = "?"
        self.master_name = "%s:%s" % (hostname, os.path.abspath(self.basedir))
        self.master_incarnation = "pid%d-boot%d" % (os.getpid(), time.time())

        self.botmaster = BotMaster()
        self.botmaster.setName("botmaster")
        self.botmaster.setMasterName(self.master_name, self.master_incarnation)
        self.botmaster.setServiceParent(self)
        self.dispatcher.botmaster = self.botmaster

        self.status = Status(self.botmaster, self.basedir)
        self.statusTargets = []

        self.db = None
        self.db_url = None
        self.db_poll_interval = _Unset
        if db_spec:
            self.loadDatabase(db_spec)

        self.readConfig = False
        
        # create log_rotation object and set default parameters (used by WebStatus)
        self.log_rotation = LogRotation()

    def startService(self):
        service.MultiService.startService(self)
        if not self.readConfig:
            # TODO: consider catching exceptions during this call to
            # loadTheConfigFile and bailing (reactor.stop) if it fails,
            # since without a config file we can't do anything except reload
            # the config file, and it would be nice for the user to discover
            # this quickly.
            self.loadTheConfigFile()
        if hasattr(signal, "SIGHUP"):
            signal.signal(signal.SIGHUP, self._handleSIGHUP)
        for b in self.botmaster.builders.values():
            b.builder_status.addPointEvent(["master", "started"])
            b.builder_status.saveYourself()

    def _handleSIGHUP(self, *args):
        reactor.callLater(0, self.loadTheConfigFile)

    def getStatus(self):
        """
        @rtype: L{buildbot.status.builder.Status}
        """
        return self.status

    def loadTheConfigFile(self, configFile=None):
        if not configFile:
            configFile = os.path.join(self.basedir, self.configFileName)

        log.msg("Creating BuildMaster -- buildbot.version: %s" % buildbot.version)
        log.msg("loading configuration from %s" % configFile)
        configFile = os.path.expanduser(configFile)

        try:
            f = open(configFile, "r")
        except IOError, e:
            log.msg("unable to open config file '%s'" % configFile)
            log.msg("leaving old configuration in place")
            log.err(e)
            return

        try:
            d = self.loadConfig(f)
        except:
            log.msg("error during loadConfig")
            log.err()
            log.msg("The new config file is unusable, so I'll ignore it.")
            log.msg("I will keep using the previous config file instead.")
            return # sorry unit tests
        f.close()
        return d # for unit tests

    def loadConfig(self, f, check_synchronously_only=False):
        """Internal function to load a specific configuration file. Any
        errors in the file will be signalled by raising an exception.

        If check_synchronously_only=True, I will return (with None)
        synchronously, after checking the config file for sanity, or raise an
        exception. I may also emit some DeprecationWarnings.

        If check_synchronously_only=False, I will return a Deferred that
        fires (with None) when the configuration changes have been completed.
        This may involve a round-trip to each buildslave that was involved."""

        localDict = {'basedir': os.path.expanduser(self.basedir)}
        try:
            exec f in localDict
        except:
            log.msg("error while parsing config file")
            raise

        try:
            config = localDict['BuildmasterConfig']
        except KeyError:
            log.err("missing config dictionary")
            log.err("config file must define BuildmasterConfig")
            raise

        known_keys = ("slaves", "change_source",
                      "schedulers", "builders", "mergeRequests",
                      "slavePortnum", "debugPassword", "logCompressionLimit",
                      "manhole", "status", "projectName", "projectURL",
                      "buildbotURL", "properties", "prioritizeBuilders",
                      "eventHorizon", "buildCacheSize", "logHorizon", "buildHorizon",
                      "changeHorizon", "logMaxSize", "logMaxTailSize",
                      "logCompressionMethod", "db_url", "multiMaster",
                      "db_poll_interval",
                      )
        for k in config.keys():
            if k not in known_keys:
                log.msg("unknown key '%s' defined in config dictionary" % k)

        try:
            # required
            schedulers = config['schedulers']
            builders = config['builders']
            slavePortnum = config['slavePortnum']
            #slaves = config['slaves']
            #change_source = config['change_source']

            # optional
            db_url = config.get("db_url", "sqlite:///state.sqlite")
            db_poll_interval = config.get("db_poll_interval", None)
            debugPassword = config.get('debugPassword')
            manhole = config.get('manhole')
            status = config.get('status', [])
            projectName = config.get('projectName')
            projectURL = config.get('projectURL')
            buildbotURL = config.get('buildbotURL')
            properties = config.get('properties', {})
            buildCacheSize = config.get('buildCacheSize', None)
            eventHorizon = config.get('eventHorizon', 50)
            logHorizon = config.get('logHorizon', None)
            buildHorizon = config.get('buildHorizon', None)
            logCompressionLimit = config.get('logCompressionLimit', 4*1024)
            if logCompressionLimit is not None and not \
                    isinstance(logCompressionLimit, int):
                raise ValueError("logCompressionLimit needs to be bool or int")
            logCompressionMethod = config.get('logCompressionMethod', "bz2")
            if logCompressionMethod not in ('bz2', 'gz'):
                raise ValueError("logCompressionMethod needs to be 'bz2', or 'gz'")
            logMaxSize = config.get('logMaxSize')
            if logMaxSize is not None and not \
                    isinstance(logMaxSize, int):
                raise ValueError("logMaxSize needs to be None or int")
            logMaxTailSize = config.get('logMaxTailSize')
            if logMaxTailSize is not None and not \
                    isinstance(logMaxTailSize, int):
                raise ValueError("logMaxTailSize needs to be None or int")
            mergeRequests = config.get('mergeRequests')
            if mergeRequests is not None and not callable(mergeRequests):
                raise ValueError("mergeRequests must be a callable")
            prioritizeBuilders = config.get('prioritizeBuilders')
            if prioritizeBuilders is not None and not callable(prioritizeBuilders):
                raise ValueError("prioritizeBuilders must be callable")
            changeHorizon = config.get("changeHorizon")
            if changeHorizon is not None and not isinstance(changeHorizon, int):
                raise ValueError("changeHorizon needs to be an int")

            multiMaster = config.get("multiMaster", False)

        except KeyError:
            log.msg("config dictionary is missing a required parameter")
            log.msg("leaving old configuration in place")
            raise

        if "sources" in config:
            m = ("c['sources'] is deprecated as of 0.7.6 and is no longer "
                 "accepted in >= 0.8.0 . Please use c['change_source'] instead.")
            raise KeyError(m)

        if "bots" in config:
            m = ("c['bots'] is deprecated as of 0.7.6 and is no longer "
                 "accepted in >= 0.8.0 . Please use c['slaves'] instead.")
            raise KeyError(m)

        slaves = config.get('slaves', [])
        if "slaves" not in config:
            log.msg("config dictionary must have a 'slaves' key")
            log.msg("leaving old configuration in place")
            raise KeyError("must have a 'slaves' key")

        if changeHorizon is not None:
            self.change_svc.changeHorizon = changeHorizon

        change_source = config.get('change_source', [])
        if isinstance(change_source, (list, tuple)):
            change_sources = change_source
        else:
            change_sources = [change_source]

        # do some validation first
        for s in slaves:
            assert interfaces.IBuildSlave.providedBy(s)
            if s.slavename in ("debug", "change", "status"):
                raise KeyError(
                    "reserved name '%s' used for a bot" % s.slavename)
        if config.has_key('interlocks'):
            raise KeyError("c['interlocks'] is no longer accepted")
        assert self.db_url is None or db_url == self.db_url, \
                "Cannot change db_url after master has started"
        assert db_poll_interval is None or isinstance(db_poll_interval, int), \
               "db_poll_interval must be an integer: seconds between polls"
        assert self.db_poll_interval is _Unset or db_poll_interval == self.db_poll_interval, \
               "Cannot change db_poll_interval after master has started"

        assert isinstance(change_sources, (list, tuple))
        for s in change_sources:
            assert interfaces.IChangeSource(s, None)
        # this assertion catches c['schedulers'] = Scheduler(), since
        # Schedulers are service.MultiServices and thus iterable.
        errmsg = "c['schedulers'] must be a list of Scheduler instances"
        assert isinstance(schedulers, (list, tuple)), errmsg
        for s in schedulers:
            assert interfaces.IScheduler(s, None), errmsg
        assert isinstance(status, (list, tuple))
        for s in status:
            assert interfaces.IStatusReceiver(s, None)

        slavenames = [s.slavename for s in slaves]
        buildernames = []
        dirnames = []

        # convert builders from objects to config dictionaries
        builders_dicts = []
        for b in builders:
            if isinstance(b, BuilderConfig):
                builders_dicts.append(b.getConfigDict())
            elif type(b) is dict:
                builders_dicts.append(b)
            else:
                raise ValueError("builder %s is not a BuilderConfig object (or a dict)" % b)
        builders = builders_dicts

        for b in builders:
            if b.has_key('slavename') and b['slavename'] not in slavenames:
                raise ValueError("builder %s uses undefined slave %s" \
                                 % (b['name'], b['slavename']))
            for n in b.get('slavenames', []):
                if n not in slavenames:
                    raise ValueError("builder %s uses undefined slave %s" \
                                     % (b['name'], n))
            if b['name'] in buildernames:
                raise ValueError("duplicate builder name %s"
                                 % b['name'])
            buildernames.append(b['name'])

            # sanity check name (BuilderConfig does this too)
            if b['name'].startswith("_"):
                errmsg = ("builder names must not start with an "
                          "underscore: " + b['name'])
                log.err(errmsg)
                raise ValueError(errmsg)

            # Fix the dictionnary with default values, in case this wasn't
            # specified with a BuilderConfig object (which sets the same defaults)
            b.setdefault('builddir', safeTranslate(b['name']))
            b.setdefault('slavebuilddir', b['builddir'])
            b.setdefault('buildHorizon', buildHorizon)
            b.setdefault('logHorizon', logHorizon)
            b.setdefault('eventHorizon', eventHorizon)
            if b['builddir'] in dirnames:
                raise ValueError("builder %s reuses builddir %s"
                                 % (b['name'], b['builddir']))
            dirnames.append(b['builddir'])

        unscheduled_buildernames = buildernames[:]
        schedulernames = []
        for s in schedulers:
            for b in s.listBuilderNames():
                # Skip checks for builders in multimaster mode
                if not multiMaster:
                    assert b in buildernames, \
                           "%s uses unknown builder %s" % (s, b)
                if b in unscheduled_buildernames:
                    unscheduled_buildernames.remove(b)

            if s.name in schedulernames:
                msg = ("Schedulers must have unique names, but "
                       "'%s' was a duplicate" % (s.name,))
                raise ValueError(msg)
            schedulernames.append(s.name)

        # Skip the checks for builders in multimaster mode
        if not multiMaster and unscheduled_buildernames:
            log.msg("Warning: some Builders have no Schedulers to drive them:"
                    " %s" % (unscheduled_buildernames,))

        # assert that all locks used by the Builds and their Steps are
        # uniquely named.
        lock_dict = {}
        for b in builders:
            for l in b.get('locks', []):
                if isinstance(l, locks.LockAccess): # User specified access to the lock
                    l = l.lockid
                if lock_dict.has_key(l.name):
                    if lock_dict[l.name] is not l:
                        raise ValueError("Two different locks (%s and %s) "
                                         "share the name %s"
                                         % (l, lock_dict[l.name], l.name))
                else:
                    lock_dict[l.name] = l
            # TODO: this will break with any BuildFactory that doesn't use a
            # .steps list, but I think the verification step is more
            # important.
            for s in b['factory'].steps:
                for l in s[1].get('locks', []):
                    if isinstance(l, locks.LockAccess): # User specified access to the lock
                        l = l.lockid
                    if lock_dict.has_key(l.name):
                        if lock_dict[l.name] is not l:
                            raise ValueError("Two different locks (%s and %s)"
                                             " share the name %s"
                                             % (l, lock_dict[l.name], l.name))
                    else:
                        lock_dict[l.name] = l

        if not isinstance(properties, dict):
            raise ValueError("c['properties'] must be a dictionary")

        # slavePortnum supposed to be a strports specification
        if type(slavePortnum) is int:
            slavePortnum = "tcp:%d" % slavePortnum

        if check_synchronously_only:
            return
        # now we're committed to implementing the new configuration, so do
        # it atomically
        # TODO: actually, this is spread across a couple of Deferreds, so it
        # really isn't atomic.

        d = defer.succeed(None)

        self.projectName = projectName
        self.projectURL = projectURL
        self.buildbotURL = buildbotURL

        self.properties = Properties()
        self.properties.update(properties, self.configFileName)

        self.status.logCompressionLimit = logCompressionLimit
        self.status.logCompressionMethod = logCompressionMethod
        self.status.logMaxSize = logMaxSize
        self.status.logMaxTailSize = logMaxTailSize
        # Update any of our existing builders with the current log parameters.
        # This is required so that the new value is picked up after a
        # reconfig.
        for builder in self.botmaster.builders.values():
            builder.builder_status.setLogCompressionLimit(logCompressionLimit)
            builder.builder_status.setLogCompressionMethod(logCompressionMethod)
            builder.builder_status.setLogMaxSize(logMaxSize)
            builder.builder_status.setLogMaxTailSize(logMaxTailSize)

        if mergeRequests is not None:
            self.botmaster.mergeRequests = mergeRequests
        if prioritizeBuilders is not None:
            self.botmaster.prioritizeBuilders = prioritizeBuilders

        self.buildCacheSize = buildCacheSize
        self.eventHorizon = eventHorizon
        self.logHorizon = logHorizon
        self.buildHorizon = buildHorizon

        # Set up the database
        d.addCallback(lambda res:
                      self.loadConfig_Database(db_url, db_poll_interval))

        # self.slaves: Disconnect any that were attached and removed from the
        # list. Update self.checker with the new list of passwords, including
        # debug/change/status.
        d.addCallback(lambda res: self.loadConfig_Slaves(slaves))

        # self.debugPassword
        if debugPassword:
            self.checker.addUser("debug", debugPassword)
            self.debugPassword = debugPassword

        # self.manhole
        if manhole != self.manhole:
            # changing
            if self.manhole:
                # disownServiceParent may return a Deferred
                d.addCallback(lambda res: self.manhole.disownServiceParent())
                def _remove(res):
                    self.manhole = None
                    return res
                d.addCallback(_remove)
            if manhole:
                def _add(res):
                    self.manhole = manhole
                    manhole.setServiceParent(self)
                d.addCallback(_add)

        # add/remove self.botmaster.builders to match builders. The
        # botmaster will handle startup/shutdown issues.
        d.addCallback(lambda res: self.loadConfig_Builders(builders))

        d.addCallback(lambda res: self.loadConfig_status(status))

        # Schedulers are added after Builders in case they start right away
        d.addCallback(lambda res:
                      self.scheduler_manager.updateSchedulers(schedulers))
        # and Sources go after Schedulers for the same reason
        d.addCallback(lambda res: self.loadConfig_Sources(change_sources))

        # self.slavePort
        if self.slavePortnum != slavePortnum:
            if self.slavePort:
                def closeSlavePort(res):
                    d1 = self.slavePort.disownServiceParent()
                    self.slavePort = None
                    return d1
                d.addCallback(closeSlavePort)
            if slavePortnum is not None:
                def openSlavePort(res):
                    self.slavePort = strports.service(slavePortnum,
                                                      self.slaveFactory)
                    self.slavePort.setServiceParent(self)
                d.addCallback(openSlavePort)
                log.msg("BuildMaster listening on port %s" % slavePortnum)
            self.slavePortnum = slavePortnum

        log.msg("configuration update started")
        def _done(res):
            self.readConfig = True
            log.msg("configuration update complete")
        d.addCallback(_done)
        d.addCallback(lambda res: self.botmaster.triggerNewBuildCheck())
        d.addErrback(log.err)
        return d

    def loadDatabase(self, db_spec, db_poll_interval=None):
        if self.db:
            return

        # make sure it's up to date
        sm = DBSchemaManager(db_spec, self.basedir)
        if not sm.is_current():
            raise exceptions.DatabaseNotReadyError, textwrap.dedent("""
                The Buildmaster database needs to be upgraded before this version of buildbot
                can run.  Use the following command-line
                    buildbot upgrade-master path/to/master
                to upgrade the database, and try starting the buildmaster again.  You may want
                to make a backup of your buildmaster before doing so.""")

        self.db = connector.DBConnector(db_spec)
        self.db.start()

        self.botmaster.db = self.db
        self.status.setDB(self.db)

        self.db.subscribe_to("add-buildrequest",
                             self.botmaster.trigger_add_buildrequest)

        sm = SchedulerManager(self, self.db, self.change_svc)
        self.db.subscribe_to("add-change", sm.trigger_add_change)
        self.db.subscribe_to("modify-buildset", sm.trigger_modify_buildset)

        self.scheduler_manager = sm
        sm.setServiceParent(self)

        # Set db_poll_interval (perhaps to 30 seconds) if you are using
        # multiple buildmasters that share a common database, such that the
        # masters need to discover what each other is doing by polling the
        # database. TODO: this will be replaced by the DBNotificationServer.
        if db_poll_interval:
            # it'd be nice if TimerService let us set now=False
            t1 = TimerService(db_poll_interval, sm.trigger)
            t1.setServiceParent(self)
            t2 = TimerService(db_poll_interval, self.botmaster.loop.trigger)
            t2.setServiceParent(self)
        # adding schedulers (like when loadConfig happens) will trigger the
        # scheduler loop at least once, which we need to jump-start things
        # like Periodic.

    def loadConfig_Database(self, db_url, db_poll_interval):
        self.db_url = db_url
        self.db_poll_interval = db_poll_interval
        db_spec = DBSpec.from_url(db_url, self.basedir)
        self.loadDatabase(db_spec, db_poll_interval)

    def loadConfig_Slaves(self, new_slaves):
        # set up the Checker with the names and passwords of all valid slaves
        self.checker.users = {} # violates abstraction, oh well
        for s in new_slaves:
            self.checker.addUser(s.slavename, s.password)
        self.checker.addUser("change", "changepw")
        # let the BotMaster take care of the rest
        return self.botmaster.loadConfig_Slaves(new_slaves)

    def loadConfig_Sources(self, sources):
        if not sources:
            log.msg("warning: no ChangeSources specified in c['change_source']")
        # shut down any that were removed, start any that were added
        deleted_sources = [s for s in self.change_svc if s not in sources]
        added_sources = [s for s in sources if s not in self.change_svc]
        log.msg("adding %d new changesources, removing %d" %
                (len(added_sources), len(deleted_sources)))
        dl = [self.change_svc.removeSource(s) for s in deleted_sources]
        def addNewOnes(res):
            [self.change_svc.addSource(s) for s in added_sources]
        d = defer.DeferredList(dl, fireOnOneErrback=1, consumeErrors=0)
        d.addCallback(addNewOnes)
        return d

    def allSchedulers(self):
        return list(self.scheduler_manager)

    def loadConfig_Builders(self, newBuilderData):
        somethingChanged = False
        newList = {}
        newBuilderNames = []
        allBuilders = self.botmaster.builders.copy()
        for data in newBuilderData:
            name = data['name']
            newList[name] = data
            newBuilderNames.append(name)

        # identify all that were removed
        for oldname in self.botmaster.getBuildernames():
            if oldname not in newList:
                log.msg("removing old builder %s" % oldname)
                del allBuilders[oldname]
                somethingChanged = True
                # announce the change
                self.status.builderRemoved(oldname)

        # everything in newList is either unchanged, changed, or new
        for name, data in newList.items():
            old = self.botmaster.builders.get(name)
            basedir = data['builddir']
            #name, slave, builddir, factory = data
            if not old: # new
                # category added after 0.6.2
                category = data.get('category', None)
                log.msg("adding new builder %s for category %s" %
                        (name, category))
                statusbag = self.status.builderAdded(name, basedir, category)
                builder = Builder(data, statusbag)
                allBuilders[name] = builder
                somethingChanged = True
            elif old.compareToSetup(data):
                # changed: try to minimize the disruption and only modify the
                # pieces that really changed
                diffs = old.compareToSetup(data)
                log.msg("updating builder %s: %s" % (name, "\n".join(diffs)))

                statusbag = old.builder_status
                statusbag.saveYourself() # seems like a good idea
                # TODO: if the basedir was changed, we probably need to make
                # a new statusbag
                new_builder = Builder(data, statusbag)
                new_builder.consumeTheSoulOfYourPredecessor(old)
                # that migrates any retained slavebuilders too

                # point out that the builder was updated. On the Waterfall,
                # this will appear just after any currently-running builds.
                statusbag.addPointEvent(["config", "updated"])

                allBuilders[name] = new_builder
                somethingChanged = True
            else:
                # unchanged: leave it alone
                log.msg("builder %s is unchanged" % name)
                pass

        # regardless of whether anything changed, get each builder status
        # to update its config
        for builder in allBuilders.values():
            builder.builder_status.reconfigFromBuildmaster(self)

        # and then tell the botmaster if anything's changed
        if somethingChanged:
            sortedAllBuilders = [allBuilders[name] for name in newBuilderNames]
            d = self.botmaster.setBuilders(sortedAllBuilders)
            return d
        return None

    def loadConfig_status(self, status):
        dl = []

        # remove old ones
        for s in self.statusTargets[:]:
            if not s in status:
                log.msg("removing IStatusReceiver", s)
                d = defer.maybeDeferred(s.disownServiceParent)
                dl.append(d)
                self.statusTargets.remove(s)
        # after those are finished going away, add new ones
        def addNewOnes(res):
            for s in status:
                if not s in self.statusTargets:
                    log.msg("adding IStatusReceiver", s)
                    s.setServiceParent(self)
                    self.statusTargets.append(s)
        d = defer.DeferredList(dl, fireOnOneErrback=1)
        d.addCallback(addNewOnes)
        return d


    def addChange(self, change):
        self.db.addChangeToDatabase(change)
        self.status.changeAdded(change)

    def triggerSlaveManager(self):
        self.botmaster.triggerNewBuildCheck()

    def submitBuildSet(self, builderNames, ss, reason, props=None, now=False):
        # determine the set of Builders to use
        for name in builderNames:
            b = self.botmaster.builders.get(name)
            if not b:
                raise KeyError("no such builder named '%s'" % name)
            if now and not b.slaves:
                raise interfaces.NoSlaveError
        if props is None:
            props = Properties()
        bsid = self.db.runInteractionNow(self._txn_submitBuildSet,
                                         builderNames, ss, reason, props)
        return BuildSetStatus(bsid, self.status, self.db)

    def _txn_submitBuildSet(self, t, builderNames, ss, reason, props):
        ssid = self.db.get_sourcestampid(ss, t)
        bsid = self.db.create_buildset(ssid, reason, props, builderNames, t)
        return bsid


class Control:
    implements(interfaces.IControl)

    def __init__(self, master):
        self.master = master

    def addChange(self, change):
        self.master.addChange(change)

    def submitBuildSet(self, builderNames, ss, reason, props=None, now=False):
        return self.master.submitBuildSet(builderNames, ss, reason, props, now)

    def getBuilder(self, name):
        b = self.master.botmaster.builders[name]
        return BuilderControl(b, self)

components.registerAdapter(Control, BuildMaster, interfaces.IControl)




import os
from zope.interface import Interface, implements
from buildbot.status.web.base import HtmlResource

class IAuth(Interface):
    """Represent an authentication method."""

    def authenticate(self, user, passwd):
            """Check whether C{user} / C{passwd} are valid."""

    def errmsg(self):
            """Get the reason authentication failed."""

class AuthBase:
    err = ""

    def errmsg(self):
        return self.err

class BasicAuth(AuthBase):
    implements(IAuth)
    """Implement basic authentication against a list of user/passwd."""

    userpass = []
    """List of user/pass tuples."""

    def __init__(self, userpass):
        """C{userpass} is a list of (user, passwd)."""
        for item in userpass:
            assert isinstance(item, tuple)
            u, p = item
            assert isinstance(u, str)
            assert isinstance(p, str)
        self.userpass = userpass

    def authenticate(self, user, passwd):
        """Check that C{user}/C{passwd} is a valid user/pass tuple."""
        if not self.userpass:
            self.err = "Bad self.userpass data"
            return False
        for u, p in self.userpass:
            if user == u and passwd == p:
                self.err = ""
                return True
        self.err = "Invalid username or password"
        return False

class HTPasswdAuth(AuthBase):
    implements(IAuth)
    """Implement authentication against an .htpasswd file."""

    file = ""
    """Path to the .htpasswd file to use."""

    def __init__(self, file):
        """C{file} is a path to an .htpasswd file."""
        assert os.path.exists(file)
        self.file = file

    def authenticate(self, user, passwd):
        """Authenticate C{user} and C{passwd} against an .htpasswd file"""
        if not os.path.exists(self.file):
            self.err = "No such file: " + self.file
            return False
        # Fetch each line from the .htpasswd file and split it into a
        # [user, passwd] array.
        lines = [l.rstrip().split(':', 1)
                 for l in file(self.file).readlines()]
        # Keep only the line for this login
        lines = [l for l in lines if l[0] == user]
        if not lines:
            self.err = "Invalid user/passwd"
            return False
        # This is the DES-hash of the password. The first two characters are
        # the salt used to introduce disorder in the DES algorithm.
        hash = lines[0][1]
        from crypt import crypt #@UnresolvedImport
        res = hash == crypt(passwd, hash[0:2])
        if res:
            self.err = ""
        else:
            self.err = "Invalid user/passwd"
        return res

class AuthFailResource(HtmlResource):
    title = "Authentication Failed"

    def content(self, request, cxt):
        template = request.site.buildbot_service.templates.get_template("authfail.html")
        return template.render(**cxt)
    




import tempfile
import os
from cStringIO import StringIO

from twisted.python import log
from twisted.application import service
from twisted.internet import defer, protocol, error, reactor
from twisted.internet.task import LoopingCall

from buildbot import util
from buildbot.interfaces import IChangeSource
from buildbot.changes.changes import Change

class _MTProtocol(protocol.ProcessProtocol):

    def __init__(self, deferred, cmdline):
        self.cmdline = cmdline
        self.deferred = deferred
        self.s = StringIO()

    def errReceived(self, text):
        log.msg("stderr: %s" % text)

    def outReceived(self, text):
        log.msg("stdout: %s" % text)
        self.s.write(text)

    def processEnded(self, reason):
        log.msg("Command %r exited with value %s" % (self.cmdline, reason))
        if isinstance(reason.value, error.ProcessDone):
            self.deferred.callback(self.s.getvalue())
        else:
            self.deferred.errback(reason)

class Monotone:
    """All methods of this class return a Deferred."""

    def __init__(self, bin, db):
        self.bin = bin
        self.db = db

    def _run_monotone(self, args):
        d = defer.Deferred()
        cmdline = (self.bin, "--db=" + self.db) + tuple(args)
        p = _MTProtocol(d, cmdline)
        log.msg("Running command: %r" % (cmdline,))
        log.msg("wd: %s" % os.getcwd())
        reactor.spawnProcess(p, self.bin, cmdline)
        return d

    def _process_revision_list(self, output):
        if output:
            return output.strip().split("\n")
        else:
            return []

    def get_interface_version(self):
        d = self._run_monotone(["automate", "interface_version"])
        d.addCallback(self._process_interface_version)
        return d

    def _process_interface_version(self, output):
        return tuple(map(int, output.strip().split(".")))

    def db_init(self):
        return self._run_monotone(["db", "init"])

    def db_migrate(self):
        return self._run_monotone(["db", "migrate"])

    def pull(self, server, pattern):
        return self._run_monotone(["pull", server, pattern])

    def get_revision(self, rid):
        return self._run_monotone(["cat", "revision", rid])

    def get_heads(self, branch, rcfile=""):
        cmd = ["automate", "heads", branch]
        if rcfile:
            cmd += ["--rcfile=" + rcfile]
        d = self._run_monotone(cmd)
        d.addCallback(self._process_revision_list)
        return d

    def erase_ancestors(self, revs):
        d = self._run_monotone(["automate", "erase_ancestors"] + revs)
        d.addCallback(self._process_revision_list)
        return d

    def ancestry_difference(self, new_rev, old_revs):
        d = self._run_monotone(["automate", "ancestry_difference", new_rev]
                               + old_revs)
        d.addCallback(self._process_revision_list)
        return d

    def descendents(self, rev):
        d = self._run_monotone(["automate", "descendents", rev])
        d.addCallback(self._process_revision_list)
        return d

    def log(self, rev, depth=None):
        if depth is not None:
            depth_arg = ["--last=%i" % (depth,)]
        else:
            depth_arg = []
        return self._run_monotone(["log", "-r", rev] + depth_arg)


class MonotoneSource(service.Service, util.ComparableMixin):
    """This source will poll a monotone server for changes and submit them to
    the change master.

    @param server_addr: monotone server specification (host:portno)

    @param branch: monotone branch to watch

    @param trusted_keys: list of keys whose code you trust

    @param db_path: path to monotone database to pull into

    @param pollinterval: interval in seconds between polls, defaults to 10 minutes
    @param monotone_exec: path to monotone executable, defaults to "monotone"
    """

    __implements__ = IChangeSource, service.Service.__implements__
    compare_attrs = ["server_addr", "trusted_keys", "db_path",
                     "pollinterval", "branch", "monotone_exec"]

    parent = None # filled in when we're added
    done_revisions = []
    last_revision = None
    loop = None
    d = None
    tmpfile = None
    monotone = None
    volatile = ["loop", "d", "tmpfile", "monotone"]

    def __init__(self, server_addr, branch, trusted_keys, db_path,
                 pollinterval=60 * 10, monotone_exec="monotone"):
        self.server_addr = server_addr
        self.branch = branch
        self.trusted_keys = trusted_keys
        self.db_path = db_path
        self.pollinterval = pollinterval
        self.monotone_exec = monotone_exec
        self.monotone = Monotone(self.monotone_exec, self.db_path)

    def startService(self):
        self.loop = LoopingCall(self.start_poll)
        self.loop.start(self.pollinterval)
        service.Service.startService(self)

    def stopService(self):
        self.loop.stop()
        return service.Service.stopService(self)

    def describe(self):
        return "monotone_source %s %s" % (self.server_addr,
                                          self.branch)

    def start_poll(self):
        if self.d is not None:
            log.msg("last poll still in progress, skipping next poll")
            return
        log.msg("starting poll")
        self.d = self._maybe_init_db()
        self.d.addCallback(self._do_netsync)
        self.d.addCallback(self._get_changes)
        self.d.addErrback(self._handle_error)

    def _handle_error(self, failure):
        log.err(failure)
        self.d = None

    def _maybe_init_db(self):
        if not os.path.exists(self.db_path):
            log.msg("init'ing db")
            return self.monotone.db_init()
        else:
            log.msg("db already exists, migrating")
            return self.monotone.db_migrate()

    def _do_netsync(self, output):
        return self.monotone.pull(self.server_addr, self.branch)

    def _get_changes(self, output):
        d = self._get_new_head()
        d.addCallback(self._process_new_head)
        return d

    def _get_new_head(self):
        # This function returns a deferred that resolves to a good pick of new
        # head (or None if there is no good new head.)

        # First need to get all new heads...
        rcfile = """function get_revision_cert_trust(signers, id, name, val)
                      local trusted_signers = { %s }
                      local ts_table = {}
                      for k, v in pairs(trusted_signers) do ts_table[v] = 1 end
                      for k, v in pairs(signers) do
                        if ts_table[v] then
                          return true
                        end
                      end
                      return false
                    end
        """
        trusted_list = ", ".join(['"' + key + '"' for key in self.trusted_keys])
        # mktemp is unsafe, but mkstemp is not 2.2 compatible.
        tmpfile_name = tempfile.mktemp()
        f = open(tmpfile_name, "w")
        f.write(rcfile % trusted_list)
        f.close()
        d = self.monotone.get_heads(self.branch, tmpfile_name)
        d.addCallback(self._find_new_head, tmpfile_name)
        return d

    def _find_new_head(self, new_heads, tmpfile_name):
        os.unlink(tmpfile_name)
        # Now get the old head's descendents...
        if self.last_revision is not None:
            d = self.monotone.descendents(self.last_revision)
        else:
            d = defer.succeed(new_heads)
        d.addCallback(self._pick_new_head, new_heads)
        return d

    def _pick_new_head(self, old_head_descendents, new_heads):
        for r in new_heads:
            if r in old_head_descendents:
                return r
        return None

    def _process_new_head(self, new_head):
        if new_head is None:
            log.msg("No new head")
            self.d = None
            return None
        # Okay, we have a new head; we need to get all the revisions since
        # then and create change objects for them.
        # Step 1: simplify set of processed revisions.
        d = self._simplify_revisions()
        # Step 2: get the list of new revisions
        d.addCallback(self._get_new_revisions, new_head)
        # Step 3: add a change for each
        d.addCallback(self._add_changes_for_revisions)
        # Step 4: all done
        d.addCallback(self._finish_changes, new_head)
        return d

    def _simplify_revisions(self):
        d = self.monotone.erase_ancestors(self.done_revisions)
        d.addCallback(self._reset_done_revisions)
        return d

    def _reset_done_revisions(self, new_done_revisions):
        self.done_revisions = new_done_revisions
        return None

    def _get_new_revisions(self, blah, new_head):
        if self.done_revisions:
            return self.monotone.ancestry_difference(new_head,
                                                     self.done_revisions)
        else:
            # Don't force feed the builder with every change since the
            # beginning of time when it's first started up.
            return defer.succeed([new_head])

    def _add_changes_for_revisions(self, revs):
        d = defer.succeed(None)
        for rid in revs:
            d.addCallback(self._add_change_for_revision, rid)
        return d

    def _add_change_for_revision(self, blah, rid):
        d = self.monotone.log(rid, 1)
        d.addCallback(self._add_change_from_log, rid)
        return d

    def _add_change_from_log(self, log, rid):
        d = self.monotone.get_revision(rid)
        d.addCallback(self._add_change_from_log_and_revision, log, rid)
        return d

    def _add_change_from_log_and_revision(self, revision, log, rid):
        # Stupid way to pull out everything inside quotes (which currently
        # uniquely identifies filenames inside a changeset).
        pieces = revision.split('"')
        files = []
        for i in range(len(pieces)):
            if (i % 2) == 1:
                files.append(pieces[i])
        # Also pull out author key and date
        author = "unknown author"
        pieces = log.split('\n')
        for p in pieces:
            if p.startswith("Author:"):
                author = p.split()[1]
        self.parent.addChange(Change(author, files, log, revision=rid))

    def _finish_changes(self, blah, new_head):
        self.done_revisions.append(new_head)
        self.last_revision = new_head
        self.d = None



import os
import shutil
import cPickle
import pprint

from twisted.trial import unittest

from buildbot.db.schema import manager
from buildbot.db import dbspec

class Thing(object):
    # simple object-with-attributes for use in faking pickled objects
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

class DBSchemaManager(unittest.TestCase):

    def setUp(self):
        self.basedir = "DBSchemaManager"
        if os.path.exists(self.basedir):
            shutil.rmtree(self.basedir)
        os.makedirs(self.basedir)

        self.spec = dbspec.DBSpec.from_url("sqlite:///state.sqlite", self.basedir)

        self.sm = manager.DBSchemaManager(self.spec, self.basedir)

    ## assertions and utils

    def assertDatabaseOKEmpty(self):
        """
        assert that the database is an upgrade of an empty db
        """
        errs = []
        c = self.spec.get_sync_connection().cursor()

        # check the version
        c.execute("SELECT * FROM version")
        if c.fetchall()[0][0] != self.sm.get_current_version():
            errs.append("VERSION is not up to date")

        # check that the remaining tables are empty
        for empty_tbl in ('changes', 'change_links', 'change_files',
                'change_properties', 'schedulers', 'scheduler_changes',
                'scheduler_upstream_buildsets', 'sourcestamps', 'patches',
                'sourcestamp_changes', 'buildsets', 'buildset_properties',
                'buildrequests', 'builds'):
            c.execute("SELECT * FROM %s" % empty_tbl)
            if len(c.fetchall()) != 0:
                errs.append("table '%s' is not empty" % empty_tbl)

        if errs:
            self.fail("; ".join(errs))

    # populate the basedir with data to test the upgrade process; this should be
    # expanded as more and more data is migrated from the basedir to the database
    def fill_basedir(self):
        self.fill_basedir_changes()

    def fill_basedir_changes(self):
        changes = [
            Thing(number=1, who='dustin', comments='hi, mom', isdir=1,
                branch=None, revision=1233, revlink='http://buildbot.net',
                when=1267419122, category=None, links=[], files=[],
                properties=Thing(properties={})),
            Thing(number=2, who='warner', comments='', isdir=0,
                branch='schedulerdb', revision=1234, revlink='http://pypi.com',
                when=1267419123, category='new', links=[], files=[],
                properties=Thing(properties={})),
            # a change with number=None should cause all changes to be renumbered
            Thing(number=None, who='catlee', comments=None, isdir=0,
                branch=None, revision=1235, revlink=None,
                when=1267419132, category=None, links=[], files=[],
                properties=Thing(properties={'name' : 'jimmy'})),
            # a change with revision=None should be ignored
            Thing(number=6, who='bhearsum', comments='', isdir=0,
                branch='fixes', revision=None, revlink=None,
                when=1267419134, category='nice', links=[], files=[],
                properties=Thing(properties={})),
            Thing(number=7, who='marcusl', comments='', isdir=0,
                branch='jinja', revision=1239, revlink=None,
                when=1267419134, category='cool', links=['http://github.com'],
                files=['main.c', 'util.c', 'ext.c'],
                properties=Thing(properties={
                    'failures' : 3,
                    'tests' : [ 'bogus1', 'bogus2', 'bogus3' ]})),
        ]

        # embed it in a Changes object and pickle it up
        changesource = Thing(changes=changes)
        f = open(os.path.join(self.basedir, "changes.pck"), "wb")
        f.write(cPickle.dumps(changesource))

    def assertDatabaseOKFull(self):
        """
        assert that the database is an upgrade of the db created by fill_basedir
        """
        errs = []
        c = self.spec.get_sync_connection().cursor()

        # check the version
        c.execute("SELECT * FROM version")
        if c.fetchall()[0][0] != self.sm.get_current_version():
            errs.append("VERSION is not up to date")

        # do a byte-for-byte comparison of the changes table and friends
        c.execute("""SELECT changeid, author, comments, is_dir, branch, revision,
                    revlink, when_timestamp, category, repository, project
                    FROM changes order by revision""")
        res = list(c.fetchall())
        if res != [
            (1, u'dustin', 'hi, mom', 1, u'', u'1233',
                u'http://buildbot.net', 1267419122, u'', u'', u''),
            (2, u'warner', u'', 0, u'schedulerdb', u'1234',
                u'http://pypi.com', 1267419123, u'new', u'', u''),
            (3, u'catlee', u'', 0, u'', u'1235',
                u'', 1267419132, u'', u'', u''),
            # note change by bhearsum is missing because its revision=None
            (4, u'marcusl', u'', 0, u'jinja', u'1239',
                u'', 1267419134, u'cool', u'', u''),
            ]:
            pprint.pprint(res)
            errs.append("changes table does not match expectations")

        c.execute("""SELECT changeid, link from change_links order by changeid""")
        res = list(c.fetchall())
        if res != [
                (4, u'http://github.com'),
            ]:
            pprint.pprint(res)
            errs.append("change_links table does not match expectations")

        c.execute("""SELECT changeid, filename from change_files order by changeid""")
        res = list(c.fetchall())
        if res != [
                (4, u'main.c'),
                (4, u'util.c'),
                (4, u'ext.c'),
            ]:
            pprint.pprint(res)
            errs.append("change_files table does not match expectations")

        c.execute("""SELECT changeid, property_name, property_value
                    from change_properties order by changeid, property_name""")
        res = list(c.fetchall())
        if res != [
                (3, u'name', u'"jimmy"'),
                (4, u'failures', u'3'),
                (4, u'tests', u'["bogus1", "bogus2", "bogus3"]'),
            ]:
            pprint.pprint(res)
            errs.append("change_properties table does not match expectations")

        # check that the remaining tables are empty
        for empty_tbl in ('schedulers', 'scheduler_changes',
                'scheduler_upstream_buildsets', 'sourcestamps', 'patches',
                'sourcestamp_changes', 'buildsets', 'buildset_properties',
                'buildrequests', 'builds'):
            c.execute("SELECT * FROM %s" % empty_tbl)
            if len(c.fetchall()) != 0:
                errs.append("table '%s' is not empty" % empty_tbl)

        if errs:
            self.fail("; ".join(errs))


    ## tests

    def test_get_current_version(self):
        # this is as much a reminder to write tests for the new version
        # as a test of the (very trivial) method
        self.assertEqual(self.sm.get_current_version(), 5)

    def test_get_db_version_empty(self):
        self.assertEqual(self.sm.get_db_version(), 0)

    def test_get_db_version_int(self):
        conn = self.spec.get_sync_connection()
        c = conn.cursor()
        c.execute("CREATE TABLE version (`version` integer)")
        c.execute("INSERT INTO version values (17)")
        self.assertEqual(self.sm.get_db_version(conn), 17)

    def test_is_current_empty(self):
        self.assertFalse(self.sm.is_current())

    def test_is_current_empty_upgrade(self):
        self.sm.upgrade(quiet=True)
        self.assertTrue(self.sm.is_current())

    def test_upgrade_empty(self):
        self.sm.upgrade(quiet=True)
        self.assertDatabaseOKEmpty()

    def test_upgrade_full(self):
        self.fill_basedir()
        self.sm.upgrade(quiet=True)
        self.assertDatabaseOKFull()

    def test_scheduler_name_uniqueness(self):
        self.sm.upgrade(quiet=True)
        c = self.spec.get_sync_connection().cursor()
        c.execute("""INSERT INTO schedulers (`name`, `class_name`, `state`)
                                             VALUES ('s1', 'Nightly', '')""")
        c.execute("""INSERT INTO schedulers (`name`, `class_name`, `state`)
                                             VALUES ('s1', 'Periodic', '')""")
        self.assertRaises(Exception, c.execute,
                """INSERT INTO schedulers (`name`, `class_name`, `state`)
                                   VALUES ('s1', 'Nightly', '')""")

class MySQLDBSchemaManager(DBSchemaManager):
    def setUp(self):
        self.basedir = "MySQLDBSchemaManager"
        if os.path.exists(self.basedir):
            shutil.rmtree(self.basedir)
        os.makedirs(self.basedir)

        self.spec = dbspec.DBSpec.from_url("mysql://buildbot_test:buildbot_test@localhost/buildbot_test")

        # Drop all previous tables
        cur = self.spec.get_sync_connection().cursor()
        cur.execute("SHOW TABLES")
        for row in cur.fetchall():
            cur.execute("DROP TABLE %s" % row[0])
        cur.execute("COMMIT")

        self.sm = manager.DBSchemaManager(self.spec, self.basedir)

try:
    import MySQLdb
    conn = MySQLdb.connect(user="buildbot_test", db="buildbot_test", passwd="buildbot_test", use_unicode=True, charset='utf8')
except:
    MySQLDBSchemaManager.skip = "MySQLdb not installed"




# This is a class which watches a maildir for new messages. It uses the
# linux dirwatcher API (if available) to look for new files. The
# .messageReceived method is invoked with the filename of the new message,
# relative to the top of the maildir (so it will look like "new/blahblah").

import os
from twisted.python import log
from twisted.application import service, internet
from twisted.internet import reactor
dnotify = None
try:
    import dnotify
except:
    # I'm not actually sure this log message gets recorded
    log.msg("unable to import dnotify, so Maildir will use polling instead")

class NoSuchMaildir(Exception):
    pass

class MaildirService(service.MultiService):
    """I watch a maildir for new messages. I should be placed as the service
    child of some MultiService instance. When running, I use the linux
    dirwatcher API (if available) or poll for new files in the 'new'
    subdirectory of my maildir path. When I discover a new message, I invoke
    my .messageReceived() method with the short filename of the new message,
    so the full name of the new file can be obtained with
    os.path.join(maildir, 'new', filename). messageReceived() should be
    overridden by a subclass to do something useful. I will not move or
    delete the file on my own: the subclass's messageReceived() should
    probably do that.
    """
    pollinterval = 10  # only used if we don't have DNotify

    def __init__(self, basedir=None):
        """Create the Maildir watcher. BASEDIR is the maildir directory (the
        one which contains new/ and tmp/)
        """
        service.MultiService.__init__(self)
        self.basedir = basedir
        self.files = []
        self.dnotify = None

    def setBasedir(self, basedir):
        # some users of MaildirService (scheduler.Try_Jobdir, in particular)
        # don't know their basedir until setServiceParent, since it is
        # relative to the buildmaster's basedir. So let them set it late. We
        # don't actually need it until our own startService.
        self.basedir = basedir

    def startService(self):
        service.MultiService.startService(self)
        self.newdir = os.path.join(self.basedir, "new")
        if not os.path.isdir(self.basedir) or not os.path.isdir(self.newdir):
            raise NoSuchMaildir("invalid maildir '%s'" % self.basedir)
        try:
            if dnotify:
                # we must hold an fd open on the directory, so we can get
                # notified when it changes.
                self.dnotify = dnotify.DNotify(self.newdir,
                                               self.dnotify_callback,
                                               [dnotify.DNotify.DN_CREATE])
        except (IOError, OverflowError):
            # IOError is probably linux<2.4.19, which doesn't support
            # dnotify. OverflowError will occur on some 64-bit machines
            # because of a python bug
            log.msg("DNotify failed, falling back to polling")
        if not self.dnotify:
            t = internet.TimerService(self.pollinterval, self.poll)
            t.setServiceParent(self)
        self.poll()

    def dnotify_callback(self):
        log.msg("dnotify noticed something, now polling")

        # give it a moment. I found that qmail had problems when the message
        # was removed from the maildir instantly. It shouldn't, that's what
        # maildirs are made for. I wasn't able to eyeball any reason for the
        # problem, and safecat didn't behave the same way, but qmail reports
        # "Temporary_error_on_maildir_delivery" (qmail-local.c:165,
        # maildir_child() process exited with rc not in 0,2,3,4). Not sure
        # why, and I'd have to hack qmail to investigate further, so it's
        # easier to just wait a second before yanking the message out of new/

        reactor.callLater(0.1, self.poll)


    def stopService(self):
        if self.dnotify:
            self.dnotify.remove()
            self.dnotify = None
        return service.MultiService.stopService(self)

    def poll(self):
        assert self.basedir
        # see what's new
        for f in self.files:
            if not os.path.isfile(os.path.join(self.newdir, f)):
                self.files.remove(f)
        newfiles = []
        for f in os.listdir(self.newdir):
            if not f in self.files:
                newfiles.append(f)
        self.files.extend(newfiles)
        # TODO: sort by ctime, then filename, since safecat uses a rather
        # fine-grained timestamp in the filename
        for n in newfiles:
            # TODO: consider catching exceptions in messageReceived
            self.messageReceived(n)

    def messageReceived(self, filename):
        """Called when a new file is noticed. Will call
        self.parent.messageReceived() with a path relative to maildir/new.
        Should probably be overridden in subclasses."""
        self.parent.messageReceived(filename)




from twisted.trial import unittest

from buildbot import util

class formatInterval(unittest.TestCase):

    def test_zero(self):
        self.assertEqual(util.formatInterval(0), "0 secs")

    def test_seconds_singular(self):
        self.assertEqual(util.formatInterval(1), "1 secs")

    def test_seconds(self):
        self.assertEqual(util.formatInterval(7), "7 secs")

    def test_minutes_one(self):
        self.assertEqual(util.formatInterval(60), "60 secs")

    def test_minutes_over_one(self):
        self.assertEqual(util.formatInterval(61), "1 mins, 1 secs")

    def test_minutes(self):
        self.assertEqual(util.formatInterval(300), "5 mins, 0 secs")

    def test_hours_one(self):
        self.assertEqual(util.formatInterval(3600), "60 mins, 0 secs")

    def test_hours_over_one_sec(self):
        self.assertEqual(util.formatInterval(3601), "1 hrs, 1 secs")

    def test_hours_over_one_min(self):
        self.assertEqual(util.formatInterval(3660), "1 hrs, 60 secs")

    def test_hours(self):
        self.assertEqual(util.formatInterval(7200), "2 hrs, 0 secs")

    def test_mixed(self):
        self.assertEqual(util.formatInterval(7392), "2 hrs, 3 mins, 12 secs")

class safeTranslate(unittest.TestCase):

    def test_str_good(self):
        self.assertEqual(util.safeTranslate(str("full")), str("full"))

    def test_str_bad(self):
        self.assertEqual(util.safeTranslate(str("speed=slow;quality=high")),
                         str("speed_slow_quality_high"))

    def test_str_pathological(self):
        # if you needed proof this wasn't for use with sensitive data
        self.assertEqual(util.safeTranslate(str("p\ath\x01ogy")),
                         str("p\ath\x01ogy")) # bad chars still here!

    def test_unicode_good(self):
        self.assertEqual(util.safeTranslate(u"full"), str("full"))

    def test_unicode_bad(self):
        self.assertEqual(util.safeTranslate(unicode("speed=slow;quality=high")),
                         str("speed_slow_quality_high"))

    def test_unicode_pathological(self):
        self.assertEqual(util.safeTranslate(u"\u0109"),
                         str("\xc4\x89")) # yuck!

class naturalSort(unittest.TestCase):

    def test_alpha(self):
        self.assertEqual(
                util.naturalSort(['x', 'aa', 'ab']),
                ['aa', 'ab', 'x'])

    def test_numeric(self):
        self.assertEqual(
                util.naturalSort(['1', '10', '11', '2', '20']),
                ['1', '2', '10', '11', '20'])

    def test_alphanum(self):
        l1 = 'aa10ab aa1ab aa10aa f a aa3 aa30 aa3a aa30a'.split()
        l2 = 'a aa1ab aa3 aa3a aa10aa aa10ab aa30 aa30a f'.split()
        self.assertEqual(util.naturalSort(l1), l2)

class LRUCache(unittest.TestCase):

    def setUp(self):
        self.lru = util.LRUCache(3)
        self.a = "AAA"
        self.b = "BBB"
        self.x = "XXX"
        self.y = "YYY"

    def test_setitem_and_getitem(self):
        self.lru['a'] = self.a
        self.assertTrue(self.lru['a'] is self.a)

    def test_full(self):
        # full, but not to overflowing
        self.lru.add("a", self.a)
        self.lru.add("x", self.x)
        self.lru.add("y", self.y)
        self.assertTrue(self.lru.get('a') is self.a)
        self.assertTrue(self.lru.get('x') is self.x)
        self.assertTrue(self.lru.get('y') is self.y)

    def test_lru_orderByAdd(self):
        # least-recently used should get kicked out
        self.lru.add("a", self.a)
        self.lru.add("b", self.b)
        self.lru.add("x", self.x)
        self.lru.add("y", self.y)
        self.assertEqual((self.lru.get('a'), self.lru.get('b')),
                         (None, self.b))

    def test_lru_usedByGet(self):
        self.lru.add("a", self.a)
        self.lru.add("b", self.b)
        self.lru.add("x", self.x)
        # use a and x, so b gets pushed out of the cache..
        self.lru.get('a')
        self.lru.get('x')
        self.lru.add("y", self.y)
        self.assertEqual((self.lru.get('a'), self.lru.get('b')),
                         (self.a, None))

    def test_lru_usedByAdd(self):
        # least-recently used should get kicked out
        self.lru.add("a", self.a)
        self.lru.add("b", self.b)
        self.lru.add("x", self.x)
        # re-add a and x, so b gets pushed out of the cache (this is a
        # regression test - older code did not mark items as used when they
        # were added)
        self.lru.add("a", self.a)
        self.lru.add("x", self.x)
        self.lru.add("y", self.y)
        self.assertEqual((self.lru.get('a'), self.lru.get('b')),
                         (self.a, None))

    def test_get_falseValue(self):
        # this more of a regression test: it used to be that getting a
        # false value would not move the key up in the LRU list
        self.lru.add("z", 0)
        self.lru.add("a", self.a)
        self.lru.add("b", self.b)
        self.lru.get("z")
        self.lru.add("x", self.x)
        self.assertEqual(self.lru.get("z"), 0)

class none_or_str(unittest.TestCase):

    def test_none(self):
        self.assertEqual(util.none_or_str(None), None)

    def test_str(self):
        self.assertEqual(util.none_or_str("hi"), "hi")

    def test_int(self):
        self.assertEqual(util.none_or_str(199), "199")



from twisted.python import reflect

# note that schema modules are not loaded unless an upgrade is taking place

CURRENT_VERSION = 5

class DBSchemaManager(object):
    """
    This class is responsible for managing the database schema and upgrading it
    as necessary.  This includes both the *actual* database and the old pickle
    database, as migrations move data between the two.

    Note that this class is *entirely synchronous*!  Performing any other operations
    while changing the schema is just asking for trouble.
    """
    def __init__(self, spec, basedir):
        self.spec = spec
        self.basedir = basedir
        self.dbapi = self.spec.get_dbapi()

    def get_db_version(self, conn=None):
        """
        Get the current schema version for this database
        """
        close_conn = False
        if not conn:
            conn = self.spec.get_sync_connection()
            close_conn = True
        c = conn.cursor()
        try:
            try:
                c.execute("SELECT version FROM version")
                rows = c.fetchall()
                assert len(rows) == 1, "%i rows in version table! (should only be 1)" % len(rows)
                return rows[0][0]
            except (self.dbapi.OperationalError, self.dbapi.ProgrammingError):
                # no version table = version 0
                return 0
        finally:
            if close_conn:
                conn.close()

    def get_current_version(self):
        """
        Get the current db version for this release of buildbot
        """
        return CURRENT_VERSION

    def is_current(self):
        """
        Is this database current?
        """
        return self.get_db_version() == self.get_current_version()

    def upgrade(self, quiet=False):
        """
        Upgrade this database to the current version
        """
        conn = self.spec.get_sync_connection()
        try:
            while self.get_db_version() < self.get_current_version():
                next_version = self.get_db_version() + 1
                next_version_module = reflect.namedModule("buildbot.db.schema.v%d" % next_version)
                upg = next_version_module.Upgrader(self.dbapi, conn, self.basedir, quiet)
                upg.upgrade()
                conn.commit()
                assert self.get_db_version() == next_version
        finally:
            conn.close()



import sys

try:
    from pysqlite2 import dbapi2 as sqlite3
except ImportError:
    # don't use built-in sqlite3 on 2.5 -- it has *bad* bugs
    if sys.version_info >= (2,6):
        import sqlite3
    else:
        raise

def get_sqlite_memory_connection():
    return sqlite3.connect(":memory:")

class FakeDBSpec(object):
    def __init__(self, conn=None, pool=None):
        self.conn = conn
        self.pool = pool

    def get_dbapi(self):
        return sqlite3

    def get_sync_connection(self):
        return self.conn

    def get_async_connection_pool(self):
        assert self.pool, "fake DBSpec will only return a pool once"
        pool = self.pool
        self.pool = None
        return pool



# -*- test-case-name: buildbot.test.test_buildstep -*-

from unittest import TestResult
from buildbot.process import buildstep

class DiscardStream:
    """A trivial thunk used to discard passthrough content."""

    def write(self, bytes):
        pass


class SubunitLogObserver(buildstep.LogLineObserver, TestResult):
    """Observe a log that may contain subunit output.

    This class extends TestResult to receive the callbacks from the subunit
    parser in the most direct fashion.
    """

    def __init__(self):
        buildstep.LogLineObserver.__init__(self)
        TestResult.__init__(self)
        try:
            from subunit import TestProtocolServer
        except ImportError:
            raise ImportError("subunit is not importable, but is required for "
                "SubunitLogObserver support.")
        self.protocol = TestProtocolServer(self, DiscardStream())

    def outLineReceived(self, line):
        """Process a received line."""
        # Impedance mismatch: subunit wants lines, observers get lines-no\n
        self.protocol.lineReceived(line + '\n')

    def startTest(self, test):
        TestResult.startTest(self, test)
        self.step.setProgress('tests', self.testsRun)

    def addError(self, test, err):
        TestResult.addError(self, test, err)
        self.issue()

    def addFailure(self, test, err):
        TestResult.addFailure(self, test, err)
        self.issue()

    def issue(self):
        """An issue - failing, erroring etc test."""
        self.step.setProgress('tests failed', len(self.failures) + len(self.errors))

# this used to be referenced here, so we keep a link for old time's sake
import buildbot.steps.subunit
SubunitShellCommand = buildbot.steps.subunit.SubunitShellCommand
